/* ------------------------- linea.c ---------------------------------------- */

/* Draw lines on SVGA. This augments the asm level routines.
 *
 * GrClearA (x, y, sx, sy, color)
 * GrPolygonA (npoints, points, color)
 *
 * These functions are pure asm and are found in subs.asm:
 *
 * GrMoveToA (x, y)
 * GrDrawToA (x, y, color)
 * GrLineA (x0, y0, x1, y1, color)
*/

#include <stdlib.h>

#ifndef USE_ASM
#define USE_ASM
#endif
#include "grh.h"


extern void far
GrClearA (uint x, uint y, uint sx, uint sy, uint color)
{
	uchar	far *va = VGA_PAGE;
	long	room;
	int	Page;

	if (!sx || !sy)
		return;

	if (!_GrWriteMode) {
		MY_OFF(va) = (uint)_GrAddr (x, y);
		room = 0x10000UL - MY_OFF(va);	/* room to end of bank */
		Page = _WritePage;
		for (; sy--; ++y) {
			if (room <= 0) {
				room += 0x10000UL;
				++Page, _GrSetPage (Page, Page);
			}
			if ((long)sx > room) {
				vgaset (MY_OFF(va), color, (uint)room);
				++Page, _GrSetPage (Page, Page);
				vgaset (0, color, sx-(uint)room);
				MY_OFF(va) = _GrSizeX - (uint)room;
				room = 0x10000UL - MY_OFF(va);
			} else {
				vgaset (MY_OFF(va), color, sx);
				MY_OFF(va) += _GrSizeX;
				room -= _GrSizeX;
			}
		}
		GrMoveToA (x+sx, y);
	} else {
		for (sx += x-1; sy--; ++y) {
			GrMoveToA (x,  y);
			GrDrawToA (sx, y, color);
		}
	}
}

extern void far
GrPolygonA (int npoints, ushort *points, uint color)
{
	uchar	far *va = VGA_PAGE;
	uint	y;			/* current y */
	int	fwdx, fwdy, bwdx, bwdy;	/* current edge tops */
	int	fdx, fdy, bdx, bdy;	/* interpolation */
	int	fx, bx;			/* segment endpoints */
	ushort	*lastpoint;		/* end of 'points' */
	ushort	*fwd, *bwd;		/* current point */
	ushort	*nfwd, *nbwd;		/* next point */
	long	room;
	uint	n;
	int	Page;

	if (npoints < 2)
		return;

	if (2 == npoints) {
		GrMoveToA (points[1], points[2]);
		GrDrawToA (points[3], points[4], color);
		return;
	}

/* locate the top vertex (use fwd as temp).
*/
	lastpoint = points + 2*(npoints - 1);
	y = (fwd = points)[1];
	for (bwd = points+2; bwd <= lastpoint; bwd += 2) {
		if (y > bwd[1])
			y = (fwd = bwd)[1];
	}

/* run through the edges, top to bottom.
*/
	nfwd = nbwd = fwd;

	if (!_GrWriteMode) {
		MY_OFF(va) = (uint)_GrAddr (0, y);
		room = 0x10000UL - MY_OFF(va);	/* room to end of bank */
		Page = _WritePage;
	}

	for (;;) {
		if (y == nfwd[1]) {
			do {
				fwd = nfwd;
				if (fwd == lastpoint)
					nfwd = points;
				else
					nfwd = fwd + 2;
			} while (y == nfwd[1]);
			if ((fdy = nfwd[1] - fwd[1]) < 0)
				break;
			fdx = nfwd[0] - (fwdx = fwd[0]);
			fwdy = 0;
		}

		if (y == nbwd[1]) {
			do {
				bwd = nbwd;
				if (bwd == points)
					nbwd = lastpoint;
				else
					nbwd = bwd - 2;
			} while (y == nbwd[1]);
			if ((bdy = nbwd[1] - bwd[1]) < 0)
				break;
			bdx = nbwd[0] - (bwdx = bwd[0]);
			bwdy = 0;
		}

		fx = fwdx + _muldiv (fwdy, fdx, fdy);
		bx = bwdx + _muldiv (bwdy, bdx, bdy);
		if (!_GrWriteMode) {
			if (fx > bx) {
				n  = fx;
				fx = bx;
				bx = n;
			}
			n = bx - fx + 1;
			if ((long)fx >= room)
				++Page, _GrSetPage (Page, Page);
			MY_OFF(va) = (uint)(0x0ffffU & (MY_OFF(va) + fx));
			room = 0x10000UL - MY_OFF(va);
			if ((long)n > room) {
				vgaset (MY_OFF(va), color, (uint)room);
				n -= (uint)room;
				MY_OFF(va) = 0;
				room = 0x10000UL;
				++Page, _GrSetPage (Page, Page);
			}
			vgaset (MY_OFF(va), color, n);
			n += _GrSizeX - bx - 1;
			MY_OFF(va) += n;
			room -= n;
		} else {
			GrMoveToA (fx, y);
			GrDrawToA (bx, y, color);
		}
		++y;
		++fwdy;
		++bwdy;
	}

/* do bottom tip/edge.
*/
	fx = fwdx + fdx;
	bx = bwdx + bdx;
	if (!_GrWriteMode) {
		if (fx > bx) {
			n  = fx;
			fx = bx;
			bx = n;
		}
		n = bx - fx + 1;
		if ((long)fx >= room)
			++Page, _GrSetPage (Page, Page);
		MY_OFF(va) = (uint)(0x0ffffU & (MY_OFF(va) + fx));
		room = 0x10000UL - MY_OFF(va);
		if ((long)n > room) {
			vgaset (MY_OFF(va), color, (uint)room);
			n -= (uint)room;
			MY_OFF(va) = 0;
			++Page, _GrSetPage (Page, Page);
		}
		vgaset (MY_OFF(va), color, n);
		GrMoveToA (bx, y);	/* restore currency */
	} else {
		GrMoveToA (fx, y);
		GrDrawToA (bx, y, color);
	}
}
