/* -------------------------- mem.c ----------------------------------------- */
/*
** Copyright (C) 1991 DJ Delorie, 24 Kirsten Ave, Rochester NH 03867-2954
**
** This file is distributed under the terms listed in the document
** "copying.dj", available from DJ Delorie at the address above.
** A copy of "copying.dj" should accompany this file; if not, a copy
** should be available from where this file was obtained.  This file
** may not be distributed without a verbatim copy of "copying.dj".
**
** This file is distributed WITHOUT ANY WARRANTY; without even the implied
** warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include <stdlib.h>
#include <string.h>

#include "grh.h"


#define	BADXY(x,y)	((x)<0 || (x)>=_GrSizeX || (y)<0 || (y)>=_GrSizeY)

#define	VIDEO_ADDR(x,y) \
	(((uint)((xy=(x)+_ActiveBase+(y)*(long)_GrSizeX) >> 16) ==_WritePage) \
	? (VGA_PAGE+(uint)xy) : _GrNewPage (xy))

static void far _GrCopyLine (int xf, int yf, int xt, int yt, int n);
static void far _GrLineGet (uchar *t, int xf, int yf, int n);
static void far _GrLinePut (int xt, int yt, int n, uchar *f);

extern void far
GrPlot (int x, int y, int c)
{
	ulong		xy;
	register uchar	far *a;

	if (BADXY (x, y))
		return;
	a = VIDEO_ADDR (x, y);
#if 0
	GrSetWriteMode (c);
#endif
	if (_GrWriteMode & GrXOR)
		*a ^= c;
	else if (_GrWriteMode & GrOR)
		*a |= c;
	else
		*a = c;
}

extern int far
GrPixel (int x, int y)
{
	ulong		xy;

	if (BADXY (x, y))
		return (0);
	 return (*VIDEO_ADDR (x, y));
}

extern void far
GrPlot3d (int x, int y, int c)
{
	ulong		xy;
	register uchar far *a;

	if (BADXY (x, y))
		return;
	a = VIDEO_ADDR (x, y);
	if ((*a < c) || (c < 0xc0))
		*a = c;
}

static void far
_GrCopyLine (int xf, int yf, int xt, int yt, int n)
{
	int	pt, pf;
	uchar	far *at, far *af;

	at = _GrAddrC (xt, yt);
	pt = _GrGetPage ();
	_GrAddrC (xt+n-1, yt);
	if (pt == _GrGetPage ()) {
		af = _GrAddrC (xf, yf);
		pf = _GrGetPage ();
		_GrAddrC (xf+n-1, yf);
		if (pf == _GrGetPage ()) {
			_GrSetPage (pf,pt);
#ifdef	USE_ASM
			vgamov ((uint)(ulong)at, (uint)(ulong)af, n);
#else
			memmove (at, af, n);
#endif
			return;
		}
	}
	while (n-- > 0)
		GrPlot (xt++, yt, GrPixel (xf++, yf));
}

extern void far
GrCopy (int xf, int yf, int xt, int yt, int w, int h)
{
	if (w <= 0 || h <= 0 ||
	    xf+w > _GrSizeX || xf < 0 || yf+h > _GrSizeY || yf < 0 ||
	    xt+w > _GrSizeX || xt < 0 || yt+h > _GrSizeY || yt < 0)
		return;

	if (yt > yf) {			/* move down */
		for (yf += h-1, yt += h-1; h > 0; --h, --yf, --yt)
			_GrCopyLine (xf, yf, xt, yt, w);
	} else {			/* move up */
		for (; h > 0; --h, ++yf, ++yt)
			_GrCopyLine (xf, yf, xt, yt, w);
	}
}

extern void far
GrMove (int xf, int yf, int xt, int yt, int w, int h, uint c)
{
	GrCopy (xf, yf, xt, yt, w, h);
	GrClear (xt, yt, w, h, c);	/* todo: handle overlap */
}

static void far
_GrLineGet (uchar *t, int xf, int yf, int n)
{
	int	pf;
	uchar	far *af;

	af = _GrAddrC (xf, yf);
	pf = _GrGetPage ();
	_GrAddrC (xf+n-1, yf);
	if (pf == _GrGetPage ()) {
#ifdef	USE_ASM
		vgaget (t, (uint)(ulong)af, n);
#else
		memmove (t, af, n);
#endif
		return;
	}
	while (n-- > 0)
		*t++ = GrPixel (xf++, yf);
}

static void far
_GrLinePut (int xt, int yt, int n, uchar *f)
{
	int	pt;
	uchar	far *at;

	at = _GrAddrC (xt, yt);
	pt = _GrGetPage ();
	_GrAddrC (xt+n-1, yt);
	if (pt == _GrGetPage ()) {
#ifdef	USE_ASM
		vgaput ((uint)(ulong)at, f, n);
#else
		memmove (at, f, n);
#endif
		return;
	}
	while (n-- > 0)
		GrPlot (xt++, yt, *f++);
}

extern void far
GrMemGet (uchar *p, int xf, int yf, int w, int h)
{
	if (w <= 0 || h <= 0 ||
	    xf+w > _GrSizeX || xf < 0 || yf+h > _GrSizeY || yf < 0)
		return;

	for (; h-- > 0; p += w)
		_GrLineGet (p, xf, yf++, w);
}

extern void far
GrMemPut (int xt, int yt, int w, int h, uchar *p)
{
	if (w <= 0 || h <= 0 ||
	    xt+w > _GrSizeX || xt < 0 || yt+h > _GrSizeY || yt < 0)
		return;

	for (; h-- > 0; p += w)
		_GrLinePut (xt, yt++, w, p);
}
