;------------------------------ subs.asm -------------------------------------
;
; NOTE: This module now uses 32bit size override and 286 style shift
; instructions.
;
;-----------------------------------------------------------------------------
;
; Some graphics routines I use regularly.
; The vga* routines take care of overlapping and also do dword operations 
; whenever possible. Don't worry, be happy.
;
; void far _SetCH (int ctype, int nbanks, int width);
; void far _SetPG (int page);
; extern uchar far * far _GrAddr (int x, int y); [ax, dx]
; int  far _SetVS (long offset);
; void far _SetWM (int mode);
;
; void far _ILoop (int pva, int pdy, int pdx, int pdvx, int pdvy, int pcolor);
; void far _SLoop (int vaddr, int lcount, int incv, int color);
;
; void far movw (int length, int fbase, int from, int tbase, int to);
;
; void far vgaput (int to, int from, int fbase, int nbytes);
;
; void far vgaget (int to, int tbase, int from, int nbytes);
;
; void far vgamov (int to, int from, int nbytes);
;
; void far vgaset (int to, int color, int nbytes);
;
; void far GrMoveToA (Uint x, Uint y);
;
; void far GrDrawToA (Uint x, Uint y, Uint color);
;
; void far GrLineA (Uint xa, Uint ya, Uint xb, Uint yb, Uint color);
;
;-----------------------------------------------------------------------------
	.286
_TEXT	SEGMENT  BYTE PUBLIC 'CODE'
_TEXT	ENDS
_DATA	SEGMENT  WORD PUBLIC 'DATA'
_DATA	ENDS
DGROUP	GROUP	_DATA
	ASSUME	CS: _TEXT, DS: DGROUP, SS: DGROUP, ES: DGROUP

VGASEG	EQU	0a000H

_DATA	SEGMENT
	EXTRN	_GrStats:WORD
	EXTRN	__ActiveBase:WORD
	EXTRN	__ReadPage:WORD
	EXTRN	__WritePage:WORD
	EXTRN	__GrBiosMode:WORD
	EXTRN	__GrType:WORD
	EXTRN	__GrSizeX:WORD
	EXTRN	__GrWriteMode:WORD

BytesPerLine	dw	200
VideoKs		dw	256		; how much memory we have (in 1KB)
FirstPage	dw	0		; 0 means VESA pages needs testing

_DATA	ENDS
_TEXT	SEGMENT

	INCLUDE macro.asm

	INCLUDE init.asm

	INCLUDE pages.asm

	INCLUDE banks.asm


;-----------------------------------------------------------------------------
;
; void far _SetCH (int ctype, int nbanks, int width);
;
;	Set the vga chip type.
;
;-----------------------------------------------------------------------------

	PUBLIC	__SetCH
__SetCH	PROC far

ctype	EQU	WORD PTR[BP+6]
cbanks	EQU	WORD PTR[BP+8]
cwidth	EQU	WORD PTR[BP+10]

	push	bp
	mov	bp,sp
	push	ax
	push	bx

	pushf
	cli
	call	[chipterm]
	popf

	mov	bx,ctype
	cmp	bx,31
	jle	__SetC1
	mov	bx,0
__SetC1:
	add	bx,bx

	mov	ax,SVGA_page[bx]
	mov	chippage,ax

	mov	ax,SVGA_bank[bx]
	mov	chipbank,ax

	mov	ax,SVGA_init[bx]
	mov	chipinit,ax

	mov	ax,SVGA_term[bx]
	mov	chipterm,ax

	mov	ax,cbanks
	mov	VideoKs,ax

	mov	ax,cwidth
	mov	BytesPerLine,ax

	pushf
	cli
	call	[chipinit]
	popf

	mov	FirstPage,0

	pop	bx
	pop	ax
	mov	sp,bp
	pop	bp
	ret

__SetCH	ENDP


;-----------------------------------------------------------------------------
;
; void far _SetPG (int page);
;
;	Set the vga active page.
;
;-----------------------------------------------------------------------------

	PUBLIC	__SetPG
__SetPG	PROC far
ppage	EQU	WORD PTR[BP+6]

	push	bp
	mov	bp,sp
	push	ax

	mov	ax,ppage
	call	setbank

	pop	ax
	mov	sp,bp
	pop	bp
	ret
__SetPG	ENDP


;-----------------------------------------------------------------------------
;
; extern uchar far * far _GrAddr (int x, int y); [ax, dx]
;
;-----------------------------------------------------------------------------

	PUBLIC	__GrAddr
__GrAddr PROC far
grx	EQU	ss:WORD PTR[BX+4]
gry	EQU	ss:WORD PTR[BX+6]

	mov	bx,sp

	mov	ax,gry
	imul	__GrSizeX
	add	ax,grx
	adc	dx,0
	add	ax,__ActiveBase
	adc	dx,__ActiveBase+2	; dx is page
	cmp	dx,__WritePage
	je	GrA1
	mov	__ReadPage,dx
	mov	__WritePage,dx
	push	ax
	mov	ax,dx
	call	setbank
	pop	ax
GrA1:
	mov	dx,VGASEG
	ret

__GrAddr ENDP

	PUBLIC	@_GrAddr
@_GrAddr PROC far

	mov	bx,ax			; bx = x
	mov	ax,dx			; ax = y
	imul	__GrSizeX
	add	ax,bx
	adc	dx,0
	add	ax,__ActiveBase		; ax = offset
	adc	dx,__ActiveBase+2	; dx = page
	cmp	dx,__WritePage
	je	GrA2
	mov	__ReadPage,dx
	mov	__WritePage,dx
	push	ax
	mov	ax,dx
	call	setbank
	pop	ax
GrA2:
	mov	dx,VGASEG
	ret

@_GrAddr ENDP


;-----------------------------------------------------------------------------
;
; int far _SetVS (long offset);
;
;	Set the vga visual start address.
;	Return 0 if extended address was set.
;
;-----------------------------------------------------------------------------

	PUBLIC	__SetVS
__SetVS	PROC far
plow	EQU	WORD PTR[BP+6]
phigh	EQU	WORD PTR[BP+8]

	push	bp
	mov	bp,sp
	push	bx
	push	cx
	push	dx
	push	si

	mov	cx,plow		; get low word
	mov	si,phigh	; get high word
	pushf
	cli
	call	[chippage]
	popf

	pop	si
	pop	dx
	pop	cx
	pop	bx
	mov	sp,bp
	pop	bp
	ret

__SetVS	ENDP


;-----------------------------------------------------------------------------
;
; void far _SetWM (int mode);
;
;	Set write mode for ILoop/SLoop.
;
;-----------------------------------------------------------------------------

	PUBLIC	__SetWM
__SetWM	PROC far
pmode	EQU	WORD PTR[BP+6]

	push	bp
	mov	bp,sp

	mov	ax,pmode
	mov	al,088H			; mov
	test	ah,003H			; GrXOR|GrOR?
	jz	$xxsm2			; no, straight mov
	test	ah,001H			; GrXOR?
	jz	$xxsm1			; no, must be GrOR
	mov	al,030H			; xor
	jmp	SHORT $xxsm2
$xxsm1:
	mov	al,008H			; or
$xxsm2:
	mov	cs:BYTE PTR $cxxx1+1,al	; set opcode
	mov	cs:BYTE PTR $cxxx2+1,al
	mov	cs:BYTE PTR $cxxx3+1,al
	mov	cs:BYTE PTR $j1xxx+1,al
	mov	cs:BYTE PTR $j2xxx+1,al
	mov	cs:BYTE PTR $j3xxx+1,al
	mov	cs:BYTE PTR $j4xxx+1,al
	mov	cs:BYTE PTR $cxxxd+1,al
	mov	cs:BYTE PTR $sxxx1+1,al
	mov	cs:BYTE PTR $sxxx2+1,al
	mov	cs:BYTE PTR $y1xxx+1,al
	mov	cs:BYTE PTR $y2xxx+1,al
	mov	cs:BYTE PTR $y3xxx+1,al
	mov	cs:BYTE PTR $y4xxx+1,al
	mov	cs:BYTE PTR $xxxx1+1,al

	mov	sp,bp
	pop	bp
	ret

__SetWM	ENDP


;-----------------------------------------------------------------------------
;
; void far _ILoop (int pva, int pdy, int pdx, int pdvx, int pdvy, int pcolor);
;
;	An implementation of Bresenham's algorithm with modifications.
;	Called from GrLine.c
;
;-----------------------------------------------------------------------------

	PUBLIC	__ILoop
__ILoop	PROC far
pva	EQU	WORD PTR[BP+6]		; si
pdy   	EQU	WORD PTR[BP+8]		; di
pdx 	EQU	WORD PTR[BP+10]		; cx (count)
pdvx  	EQU	WORD PTR[BP+12]		; dx
pdvy	EQU	WORD PTR[BP+14]		; immed
pcolor	EQU	WORD PTR[BP+16]		; ah
					; al = vga page
					; bx = err
; Modified instructions:
; x4		put constants in immediates
; xxx		mov/xor/or color type (see __setWM)
;
; Note that here we do:	new_dx = 010000H
;			new_dy = dy*new_dx/dx
;
	push	bp
	mov	bp,sp
	push	di
	push	si
	push	es

	mov	cx,pdx			; count: Rcx
	cmp	cx,16
	ja	$jloop			; long  line
	jmp	$zloop			; short line		386	486
$jloop:
	mov	dx,pdy			; dy*0x10000L		  4	  1
	xor	ax,ax			;			  2	  1
	div	cx			; new dy =dy*010000L/dx	 22	 24
	mov	di,ax			; new dy: Rdi		  2	  1
;								 ==	 ==
;								 30	 27
	mov	dx,pdvx			; dvx: Rdx
	mov	ax,__WritePage		; ppage: Ral
	mov	bx,pcolor
	mov	ah,bl			; color: Rah
	mov	bx,VGASEG
	mov	es,bx			; VGABASE: Res
	mov	bx,08000H		; err: Rbx

	mov	si,pdvy
	add	si,dx
	js	$jyneg
	or	dx,dx
	js	$jxneg
	mov	cs:WORD PTR $j1lx4+2,si	; dvy+dvx: Immed
	mov	si,pva			; va: Rsi
	jmp	$j1ll0
$jxneg:
	mov	cs:WORD PTR $j2lx4+2,si	; dvy+dvx: Immed
	mov	si,pva			; va: Rsi
	jmp	$j2ll0
$jyneg:
	or	dx,dx
	js	$jxneg1
	mov	cs:WORD PTR $j3lx4+2,si	; dvy+dvx: Immed
	mov	si,pva			; va: Rsi
	jmp	$j3ll0
$jxneg1:
	mov	cs:WORD PTR $j4lx4+2,si	; dvy+dvx: Immed
	mov	si,pva			; va: Rsi
	jmp	$j4ll0
;
; the main drawing loop (long line)
;
;
; dvx > 0  dvx+dvy > 0
;								386	486
;								===	===
$j1ll3a:
	add	si,dx			; va += dvx		  2	  1
	jc	$j1lp1			; if page crossing	  3/[7]	  1/[3]
$j1ll3:
$j1xxx:	mov	es:[si],ah		; vram = color (^=,|=)	4	2
$j1ll0:
	dec	cx			;			2	1
	js	$jlret			; if end of line	3/[7]	1/[3]
	add	bx,di			; err += dy		2	1
	jnc	$j1ll3a			; if err < dx		3/7	1/3
$j1lx4:	add	si,01234H		; va += dvx+dvy		2	1
	jnc	$j1ll3			; if NO page crossing	7/[3]	3/[1]
;								=====	=====
;								23/23	10/10
	inc	al			; ++ page
	call	setbank
	jmp	SHORT $j1ll3
$j1lp1:
	inc	al			; ++ page
	call	setbank
	jmp	SHORT $j1ll3
;
;
; dvx < 0  dvx+dvy > 0
;								386	486
;								===	===
$j2ll3a:
	add	si,dx			; va += dvx		  2	  1
	jnc	$j2lp1			; if page crossing	  3/[7]	  1/[3]
$j2ll3:
$j2xxx:	mov	es:[si],ah		; vram = color (^=,|=)	4	2
$j2ll0:
	dec	cx			;			2	1
	js	$jlret			; if end of line	3/[7]	1/[3]
	add	bx,di			; err += dy		2	1
	jnc	$j2ll3a			; if err < dx		3/7	1/3
$j2lx4:	add	si,01234H		; va += dvx+dvy		2	1
	jnc	$j2ll3			; if NO page crossing	7/[3]	3/[1]
;								=====	=====
;								23/23	10/10
	inc	al			; ++ page
	call	setbank
	jmp	SHORT $j2ll3
$j2lp1:
	dec	al			; ++ page
	call	setbank
	jmp	SHORT $j2ll3
;
;
; dvx > 0  dvx+dvy < 0
;								386	486
;								===	===
$j3ll3a:
	add	si,dx			; va += dvx		  2	  1
	jc	$j3lp1			; if page crossing	  3/[7]	  1/[3]
$j3ll3:
$j3xxx:	mov	es:[si],ah		; vram = color (^=,|=)	4	2
$j3ll0:
	dec	cx			;			2	1
	js	$jlret			; if end of line	3/[7]	1/[3]
	add	bx,di			; err += dy		2	1
	jnc	$j3ll3a			; if err < dx		3/7	1/3
$j3lx4:	add	si,01234H		; va += dvx+dvy		2	1
	jc	$j3ll3			; if NO page crossing	7/[3]	3/[1]
;								=====	=====
;								23/23	10/10
	dec	al			; ++ page
	call	setbank
	jmp	SHORT $j3ll3
$j3lp1:
	inc	al			; ++ page
	call	setbank
	jmp	SHORT $j3ll3
;
;
; dvx < 0  dvx+dvy < 0
;								386	486
;								===	===
$j4ll3a:
	add	si,dx			; va += dvx		  2	  1
	jnc	$j4lp1			; if page crossing	  3/[7]	  1/[3]
$j4ll3:
$j4xxx:	mov	es:[si],ah		; vram = color (^=,|=)	4	2
$j4ll0:
	dec	cx			;			2	1
	js	$jlret			; if end of line	3/[7]	1/[3]
	add	bx,di			; err += dy		2	1
	jnc	$j4ll3a			; if err < dx		3/7	1/3
$j4lx4:	add	si,01234H		; va += dvx+dvy		2	1
	jc	$j4ll3			; if NO page crossing	7/[3]	3/[1]
;								=====	=====
;								23/23	10/10
	dec	al			; ++ page
	call	setbank
	jmp	SHORT $j4ll3
$j4lp1:
	dec	al			; ++ page
	call	setbank
	jmp	SHORT $j4ll3
;
$jlret:
	and	ax,000ffH		; save page
	mov	__ReadPage,ax
	mov	__WritePage,ax
	mov	ax,si			; return va
	pop	es
	pop	si
	pop	di
	mov	sp,bp
	pop	bp
	ret
;
; Short line version.
;
$zloop:
	mov	ax,pdy			; 4  1 (old)
	mov	cs:WORD PTR $zlx8+2,ax	; 2  1 (old)	; dy: Immed
	mov	di,cx			; 2  1 (old)	; dx: Rdi
;					;== ==
;					; 8  3 (old)
;
	mov	dx,pdvx			; dvx: Rdx
	or	dx,dx			; pdvx ?
	mov	bx,07204H		; jc:add
	jns	$zlpdvx
	mov	bx,0732cH		; jnc:sub
$zlpdvx:
	mov	cs:BYTE PTR $zlx2,bh	; set jc/jnc
	mov	cs:BYTE PTR $zlx3,bl	; set add/sub
;
	mov	ax,pdvy
	add	ax,dx
	mov	cs:WORD PTR $zlx4+2,ax	; dvx+dvy: Immed
;	or	ax,ax			; pdvy ?
	mov	bx,07304H		; jnc:add
	jns	$zlpdvy
	mov	bx,0722cH		; jc:sub
$zlpdvy:
	mov	cs:BYTE PTR $zlx5,bh	; set jnc/jc
	mov	cs:BYTE PTR $zlx6,bl	; set add/sub
;
	mov	ax,__WritePage		; ppage: Ral
;
	mov	bx,pcolor
	mov	ah,bl			; color: Rah
;
	mov	bx,VGASEG
	mov	es,bx			; VGABASE: Res
;
	mov	bx,cx
	sar	bx,1			; err: Rbx
;
	mov	si,pva			; va: Rsi
	jmp	SHORT $zll0
;
; the main drawing loop (short line)
;
;				  386  486
;				  ===  ===
$zll3a:
	add	si,dx		;   2    1	; va += dvx
$zlx2:	jc	$zlp1		;   3\7  1\3	; if page crossing
$zlp1r:
$zlp2r:
$zll3:
$cxxx3:	mov	es:[si],ah	; 4    2	; vram = color (or ^= or |=)
$zll0:
	dec	cx		; 2    1
	js	$jlret		; 3\7  1\3	; if end of line
$zlx8:	add	bx,01234H	; 2    1 	; err += dy
	cmp	bx,di		; 2    1 	; err : dx
	jl	$zll3a		; 3/7  1/3	; err < dx
	sub	bx,di		; 2    1	; err -= dx
$zlx4:	add	si,01234H	; 2    1	; va += dvx+dvy
$zlx5:	jnc	$zll3		; 7\3  3\1	; if no page crossing
;				;===== =====
;				;27/25 12/11	; total
$zlp2:
$zlx6:	add	al,001H				; ++ page (or --)
	call	setbank
	jmp	SHORT $zlp2r
$zlp1:
$zlx3:	add	al,001H				; ++ page (or --)
	call	setbank
	jmp	SHORT $zlp1r

__ILoop	ENDP


;-----------------------------------------------------------------------------
;
; void far _SLoop (int vaddr, int lcount, int incv, int color);
;
;	Used to draw simple lines (0 and 45 degrees in 4 quadrants).
;	Called from GrLine.c
;
;-----------------------------------------------------------------------------

	PUBLIC	__SLoop
__SLoop	PROC far
vaddr	EQU	WORD PTR[BP+6]		; es:si
lcount	EQU	WORD PTR[BP+8]		; dx
incv	EQU	WORD PTR[BP+10]		; cx
color	EQU	WORD PTR[BP+12]		; bx

	push	bp
	mov	bp,sp
	push	si
	push	es

	mov	ax,__WritePage
	mov	bx,color		; color
	mov	cx,incv			; incv
	mov	dx,lcount		; lcount
	mov	si,VGASEG
	mov	es,si
	mov	si,vaddr		; vaddr

	dec	dx
	js	$slret			; nothing to do
	or	cx,cx			; incv ?
	js	$sldec			; decrementing
	jmp	SHORT $slinc		; incrementing
;
; the main drawing loop (incrementing)
;								386	486
$sll1:					;			===	===
$cxxx1:	mov	es:[si],bl		; vram = color (^=,|=)	4	2
	dec	dx			; --count		2	1
	js	$slret			;         < 0		3/[7]	1/[3]
$slinc:
	add	si,cx			; vaddr += incv		2	1
	jnc	$sll1			; vaddr out of page	7/[1]	3/[1]
;								=====	=====
;								18	8
	inc	al			; ++page
	call	setbank
	jmp	SHORT $sll1
;
; the main drawing loop (decrementing)
;
$sll2:
$cxxx2:	mov	es:[si],bl		; vram  = color (mov/xor/or)
	dec	dx			; --count
	js	$slret			;         < 0
$sldec:
	add	si,cx			; vaddr += incv
	jc	$sll2			; vaddr out of page
;
	dec	al			; --page
	call	setbank
	jmp	SHORT $sll2
;
;
$slret:
	and	ax,000ffH		; return page
	mov	__ReadPage,ax
	mov	__WritePage,ax
	mov	ax,si			; return va
	pop	es
	pop	si
	mov	sp,bp
	pop	bp
	ret

__SLoop	ENDP


;-----------------------------------------------------------------------------
;
; void far movw (int length, int fbase, int from, int tbase, int to);
;
;	The subroutine will move length words from fbase:from to tbase:to.
;	Not used much (if at all).
;
;-----------------------------------------------------------------------------

	PUBLIC _movw
_movw	PROC  far
plength EQU	WORD PTR[BP+6]
pfbase	EQU	WORD PTR[BP+8]
pfrom	EQU	WORD PTR[BP+10]
ptbase	EQU	WORD PTR[BP+12]
pto	EQU	WORD PTR[BP+14]

	push	bp
	mov	bp,sp
	push	ds
	push	es
	push	si
	push	di

	mov	es,ptbase
	mov	di,pto
	mov	ds,pfbase
	mov	si,pfrom
	mov	cx,plength

	cld
	rep	movsw

	pop	di
	pop	si
	pop	es
	pop	ds
	pop	bp
	ret

_movw	ENDP


;-----------------------------------------------------------------------------
;
; void far vgaput (int to, int from, int fbase, int nbytes);
;
;    The subroutine will move 'nbytes' bytes from fbase:from to VGASEG:to
;
;-----------------------------------------------------------------------------

	PUBLIC _vgaput
_vgaput	PROC  far
pto	EQU	WORD PTR[BP+6]
pfrom	EQU	WORD PTR[BP+8]
pfbase	EQU	WORD PTR[BP+10]
plength EQU	WORD PTR[BP+12]

	push	bp
	mov	bp,sp
	push	ds
	push	es
	push	si
	push	di

	mov	ax,VGASEG
	mov	es,ax
	mov	di,pto
	mov	ds,pfbase
	mov	si,pfrom
	mov	bx,plength

	cld
	mov	cx,di
	and	cx,03H		; head size
	sub	bx,cx		; update length
	rep	movsb		; move head
	mov	cx,bx
	shr	cx,2
;	shr	cx,1

	db	066H		; opsiz
	rep	movsb		; move body
	and	bx,03H		; tail length
	mov	cx,bx		; move tail
	rep	movsb
vgapret:
	pop	di
	pop	si
	pop	es
	pop	ds
	pop	bp
	ret

_vgaput	ENDP


;-----------------------------------------------------------------------------
;
; void far vgaget (int to, int tbase, int from, int nbytes);
;
;    The subroutine will move 'nbytes' bytes to tbase:to from VGASEG:from
;
;-----------------------------------------------------------------------------

	PUBLIC _vgaget
_vgaget	PROC  far
pto	EQU	WORD PTR[BP+6]
ptbase	EQU	WORD PTR[BP+8]
pfrom	EQU	WORD PTR[BP+10]
plength EQU	WORD PTR[BP+12]

	push	bp
	mov	bp,sp
	push	ds
	push	es
	push	si
	push	di

	mov	es,ptbase
	mov	di,pto
	mov	ax,VGASEG
	mov	ds,ax
	mov	si,pfrom
	mov	bx,plength

	cld
	mov	cx,si
	and	cx,03H		; head size
	sub	bx,cx		; update length
	rep	movsb		; move head
	mov	cx,bx
	shr	cx,2
;	shr	cx,1
;	rep	movsd		; move body
	DB	0F3H,066H,0A5H	; rep opsiz:movsd
	and	bx,03H		; tail length
	mov	cx,bx		; move tail
	rep	movsb
vgagret:
	pop	di
	pop	si
	pop	es
	pop	ds
	pop	bp
	ret

_vgaget	ENDP


;-----------------------------------------------------------------------------
;
; void far vgamov (int to, int from, int nbytes);
;
;    The subroutine will move 'nbytes' bytes to VGASEG:to from VGASEG:from
;
;-----------------------------------------------------------------------------

	PUBLIC _vgamov
_vgamov	PROC  far
pto	EQU	WORD PTR[BP+6]
pfrom	EQU	WORD PTR[BP+8]
plength EQU	WORD PTR[BP+10]

	push	bp
	mov	bp,sp
	push	ds
	push	es
	push	si
	push	di

	mov	ax,VGASEG
	mov	es,ax
	mov	di,pto
	mov	ds,ax
	mov	si,pfrom
	mov	bx,plength

	mov	ax,si		; from
	sub	ax,di		; from - to
	cmp	ax,8		; (from-to) ? 8
	jl	movmr		;           >

	cld
	mov	cx,di
	and	cx,03H		; head size
	sub	bx,cx		; update length
	rep	movsb		; move head
	mov	cx,bx
	shr	cx,2
;	shr	cx,1
;	rep	movsd		; move body
	DB	0F3H,066H,0A5H	; rep opsiz:movsd
	and	bx,03H		; tail length
	mov	cx,bx		; move tail
	rep	movsb
vgamret:
	pop	di
	pop	si
	pop	es
	pop	ds
	pop	bp
	ret
movmr:
	std
	add	si,bx		; point at last byte
	dec	si
	add	di,bx
	dec	di

	mov	cx,di
	and	cx,03H		; head size
	sub	bx,cx		; update length
	rep	movsb		; move head
	mov	cx,bx
	shr	cx,2
;	shr	cx,1
	sub	si,3		; point at last dword
	sub	di,3
;	rep	movsd		; move body
	DB	0F3H,066H,0A5H	; rep opsiz:movsd
	and	bx,03H		; tail length
	add	si,3		; point at last byte
	add	di,3
	mov	cx,bx		; move tail
	rep	movsb
	jmp	vgamret

_vgamov	ENDP


;-----------------------------------------------------------------------------
;
; void far vgaset (int to, int color, int nbytes); [di ax bx]
;
;    The subroutine will set 'nbytes' bytes at 'to' to 'color'
;    It respects the GrXOR/GrOR bits.
;    This is a local version. 'es' must be set in advance.
;
;-----------------------------------------------------------------------------

vgaset	PROC  near

	test	ah,003H		; GrXOR|GrOR?
	jnz	vgaset1		; yes
;
; This is a 32 bit version
;
vgas32:
	cmp	bx,6
	jl	vgas32x

	cld

	mov	ah,al		; for dword set
	mov	cx,ax
	db	66H		; operand size prefix
	shl	ax,16
	mov	ax,cx		; for dword set

	mov	cx,di		; 'to'
	neg	cx
	and	cx,03H		; head size
	sub	bx,cx		; update length
	rep	stosb		; set head

	mov	cx,bx		; body length
	shr	cx,2
	db	66H		; operand size prefix
	rep	stosw		; set body

	and	bx,03H		; tail length
vgas32x:
	mov	cx,bx
	rep	stosb		; set tail

	ret
;
; This is a 16 bit version
;
vgas16:
	mov	ah,al		; for word set
	cld

	mov	cx,di
	and	cx,01H		; head size
	sub	bx,cx		; update length
	rep	stosb		; set head

	mov	cx,bx
	shr	cx,1
	rep	stosw		; set body

	and	bx,01H		; tail length
	mov	cx,bx
	rep	stosb		; set tail
vgasret:
	ret

vgaset1:
	test	ah,001H		; GrXOR?
	jnz	vgaxor		; yes
vgaor:
	mov	cx,di
	and	cx,01H		; head size
	jz	vgaor1
	or	es:[di],al	; set head
	inc	di		; advance
	dec	bx		; update length
vgaor1:
	mov	cx,bx
	shr	cx,1
	jz	vgaor3
	mov	ah,al		; for word set
vgaor2:
	or	es:[di],ax	; set body
	add	di,2		; advance
	dec	cx		; loop
	jnz	vgaor2
vgaor3:
	and	bx,01H		; tail length
	jz	vgasret
	or	es:[di],al	; set tail
	ret

vgaxor:
	mov	cx,di
	and	cx,01H		; head size
	jz	vgaxor1
	xor	es:[di],al	; set head
	inc	di		; advance
	dec	bx		; update length
vgaxor1:
	mov	cx,bx
	shr	cx,1
	jz	vgaxor3
	mov	ah,al		; for word set
vgaxor2:
	xor	es:[di],ax	; set body
	add	di,2		; advance
	dec	cx		; loop
	jnz	vgaxor2
vgaxor3:
	and	bx,01H		; tail length
	jz	vgasret
	xor	es:[di],al	; set tail
	ret

vgaset	ENDP

;-----------------------------------------------------------------------------
;
; void far vgaset (int to, int color, int nbytes);
;
;    The subroutine will set 'nbytes' bytes at 'to' to 'color'
;    It respects the GrXOR/GrOR bits.
;    It actually calls the near 'vgaset' to do the job.
;
;-----------------------------------------------------------------------------

	PUBLIC _vgaset
_vgaset	PROC  far
pto	EQU	WORD PTR[BP+6]
pcolor	EQU	WORD PTR[BP+8]
plength EQU	WORD PTR[BP+10]

	push	bp
	mov	bp,sp
	push	es
	push	di

	mov	ax,VGASEG
	mov	es,ax

	mov	di,pto
	mov	ax,pcolor
	mov	bx,plength
	call	vgaset

	pop	di
	pop	es
	pop	bp
	ret

_vgaset	ENDP


x1	dw	0
y1	dw	0
va	dw	0
xydone	dw	0

;-----------------------------------------------------------------------------
;
; void far GrMoveToA (Uint x, Uint y) [ax, dx]
;
;-----------------------------------------------------------------------------

	PUBLIC	_GrMoveToA
_GrMoveToA PROC far
mtx	EQU	ss:WORD PTR[BX+4]
mty	EQU	ss:WORD PTR[BX+6]

	mov	bx,sp

	mov	ax,mtx
	mov	cs:x1,ax
	mov	ax,mty
	mov	cs:y1,ax
	mov	cs:xydone,0

	ret

_GrMoveToA ENDP

	PUBLIC	@GrMoveToA
@GrMoveToA PROC far

	mov	cs:x1,ax
	mov	cs:y1,dx
	mov	cs:xydone,0

	ret

@GrMoveToA ENDP


;-----------------------------------------------------------------------------
;
; void far GrDrawToA (Uint x, Uint y, Uint color) [ax, dx, bx]
;
;-----------------------------------------------------------------------------

	PUBLIC	_GrDrawToA
_GrDrawToA PROC far
dtx2	EQU	WORD PTR[BP+6]
dty2	EQU	WORD PTR[BP+8]
dtc	EQU	WORD PTR[BP+10]
ddx	EQU	WORD PTR[BP-2]
ddy	EQU	WORD PTR[BP-4]
svx	EQU	WORD PTR[BP-6]
svy	EQU	WORD PTR[BP-8]
vb	EQU	WORD PTR[BP-10]

	push	bp
	mov	bp,sp
	sub	sp,10
	push	di
	push	si
	push	es

	mov	ax,VGASEG
	mov	es,ax			; VGA segment

;
; Draw first point if not already done.
;
	mov	cx,0	
	cmp	cs:xydone,0
	jne	GrDT1
	inc	cx
	mov	cs:xydone,1
	mov	ax,cs:y1
	imul	__GrSizeX
	add	ax,cs:x1
	adc	dx,0
	add	ax,__ActiveBase
	adc	dx,__ActiveBase+2	; new page
	mov	cs:va,ax		; new offset
	cmp	dx,__WritePage
	je	GrDT2			; same page
	mov	__ReadPage,dx
	mov	__WritePage,dx
	mov	ax,dx
	call	setbank
GrDT2:
	mov	ax,dtc			; color
	mov	si,cs:va		; offset
$cxxxd:	mov	es:[si],al		; vram  = color (mov/xor/or)
GrDT1:
;
; get absolute x and y extents.
;
	mov	bx,1			; bx = svx
	mov	ax,dtx2			; ax = dx
	sub	ax,cs:x1
	jns	GrDT3
	neg	ax
	neg	bx
GrDT3:
	mov	ddx,ax
	mov	svx,bx

	mov	bx,__GrSizeX
	mov	ax,dty2
	sub	ax,cs:y1
	jns	GrDT4
	neg	ax
	neg	bx
GrDT4:
	mov	ddy,ax
	mov	svy,bx
;
; sort line type out as dx > dy, dx < dy or dx = dy
;
	mov	ax,ddx
	cmp	ax,ddy
	jl	GrDTl
	jg	GrDTg
;
; dx = dy: use the SimpleLoop method if length is not zero.
;
	mov	bx,ax			; stats
	add	bx,cx
	shl	bx,2
;	shl	bx,1
	add	_GrStats[bx],1
	adc	_GrStats[bx+2],0
	or	ax,ax
	jnz	GrDTz1
	jmp	GrDTx1
GrDTz1:
	mov	cx,svx			; simple line
	add	cx,svy
	mov	dx,ddx
	jmp	Sloop
;
; dx < dy: either use SimpleLoop (vertical) or InnerLoop (others)
;
GrDTl:					; dx < dy
	mov	bx,ddy
	add	bx,cx
	shl	bx,2
;	shl	bx,1
	add	_GrStats[bx],1
	adc	_GrStats[bx+2],0
	or	ax,ax
	jz	GrDTv
	mov	si,svx
	mov	bx,svy
	mov	cx,ddy
	mov	dx,ddx
	jmp	ILoop
GrDTv:
	mov	cx,svy			; vertical line
	mov	dx,ddy
	jmp	Sloop
;
; dx > dy: horizontal lines can either use 'vgaset' (memory fill) if they
; fit inside one memory bank or SimpleLoop if they don't.
; Other lines use InnerLoop.
;
GrDTg:					; dx > dy
	mov	bx,ax
	add	bx,cx
	shl	bx,2
;	shl	bx,1
	add	_GrStats[bx],1
	adc	_GrStats[bx+2],0
	cmp	ddy,0
	je	GrDTh
	mov	si,svy
	mov	bx,svx
	mov	cx,ddx
	mov	dx,ddy
	jmp	ILoop
;
; Horizontal line, account for bank crossing but move in bulk.
;
GrDTh:
;	cmp	ddx,16
;	jg	GrDThl
;
; short horizontal, use SimpleLoop
;
;	mov	cx,svx
;	mov	dx,ddx
;	jmp	Sloop

GrDThl:
	mov	bx,__GrWriteMode	; add mode to color
	or	dtc,bx
	mov	bx,cs:va
	cmp	svx,0			; check direction
	jl	GrDThn
;
; Drawing low->high addresses
;
	mov	ax,bx
	inc	bx			; first pixel
	add	ax,ddx			; ax = endpoint offset
	mov	cs:va,ax		; set end address
	jc	GrDTh1			; bank crossing
;
; low->high inside one bank
;
	mov	di,bx			; address
	mov	ax,dtc			; color
	mov	bx,ddx			; count
	call	vgaset
	jmp	GrDTx1
;
; low->high with bank crossing
;
GrDTh1:
	neg	bx			; count
	jz	GrDTh2			; nothing to move here
	mov	di,bx
	neg	di			; address
	mov	ax,dtc			; color
	call	vgaset
GrDTh2:
	mov	ax,__ReadPage
	inc	ax
	mov	__ReadPage,ax		; set new bank
	mov	__WritePage,ax
	call	setbank

	xor	di,di			; address
	mov	ax,dtc			; color
	mov	bx,cs:va		;
	inc	bx			; count
	call	vgaset
	jmp	GrDTx1
;
; Drawing high->low addresses
;
GrDThn:
	mov	ax,bx
	sub	ax,ddx			; ax = endpoint address
	mov	cs:va,ax
	jc	GrDTh3			; bank crossing
;
; high->low inside one bank
;
	mov	di,ax			; address
	mov	ax,dtc			; color
	mov	bx,ddx			; count
	call	vgaset
	jmp	GrDTx1
;
; high->low with bank crossing
;
GrDTh3:
	or	bx,bx			; bx = count
	jz	GrDTh4			; nothing this bank
	xor	di,di			; address
	mov	ax,dtc			; color
	call	vgaset
GrDTh4:
	mov	ax,__ReadPage
	dec	ax
	mov	__ReadPage,ax		; set new bank
	mov	__WritePage,ax
	call	setbank

	mov	di,cs:va		; address
	mov	ax,dtc			; color
	mov	bx,di
	neg	bx			; count
	call	vgaset
	jmp	GrDTx1
;
; Exit points.
;
GrDTx2:
	and	ax,000ffH		; save page
	mov	__ReadPage,ax
	mov	__WritePage,ax
	mov	ax,si			; save va
GrDTx:
	mov	cs:va,ax
GrDTx1:
	mov	ax,dtx2
	mov	cs:x1,ax
	mov	ax,dty2
	mov	cs:y1,ax

	pop	es
	pop	si
	pop	di
	mov	sp,bp
	pop	bp
	ret


;-----------------------------------------------------------------------------
;
; Sloop
;
;	Used to draw simple lines (0 and 45 degrees in 4 quadrants).
;	Called from GrDrawToA.
;
;-----------------------------------------------------------------------------

Sloop:
	mov	ax,__WritePage
	mov	bx,dtc			; color
	mov	si,cs:va		; vaddr
;
	dec	dx
	js	$sxret			; nothing to do
;
	or	cx,cx			; incv ?
	js	$sxdec			; decrementing
	jmp	SHORT $sxinc		; incrementing
;
; the main drawing loop (incrementing)
;								386	486
$sxl1:					;			===	===
$sxxx1:	mov	es:[si],bl		; vram = color (^=,|=)	4	2
	dec	dx			; --count		2	1
	js	$sxret			;         < 0		3/[7]	1/[3]
$sxinc:
	add	si,cx			; vaddr += incv		2	1
	jnc	$sxl1			; vaddr out of page	7/[1]	3/[1]
;								=====	=====
;								18	8
	inc	al			; ++page
	call	setbank
	jmp	SHORT $sxl1
;
; the main drawing loop (decrementing)
;
$sxl2:
$sxxx2:	mov	es:[si],bl		; vram  = color (mov/xor/or)
	dec	dx			; --count
	js	$sxret			;         < 0
$sxdec:
	add	si,cx			; vaddr += incv
	jc	$sxl2			; vaddr out of page
;
	dec	al			; --page
	call	setbank
	jmp	SHORT $sxl2

$sxret:
	jmp	GrDTx2


;-----------------------------------------------------------------------------
;
; ILoop
;
;	Used to draw general lines.
;	Called from GrDrawToA.
;
;-----------------------------------------------------------------------------

ILoop:

;pdy   	dx
;pdx 	cx (count)
;pdvx  	bx
;pdvy	si
;
; Modified instructions:
; x4		put constants in immediates
; xxx		mov/xor/or color type (see __setWM)
;
; Note that here we do:	new_dx = 010000H
;			new_dy = dy*new_dx/dx

;	cmp	cx,16
;	ja	$yloop			; long  line
;	jmp	$xloop			; short line
$yloop:					;			386	486
	xor	ax,ax			;			  2	  1
	div	cx			; new dy =dy*010000L/dx	 22	 24
	mov	di,ax			; new dy: Rdi		  2	  1
;								 ==	 ==
;								 30	 27
	mov	dx,bx			; dvx: Rdx
	mov	ax,__WritePage		; ppage: Ral
	mov	bx,dtc
	mov	ah,bl			; color: Rah
	mov	bx,08000H		; err: Rbx
;
	add	si,dx
	js	$yyneg
	or	dx,dx
	js	$yxneg
	mov	cs:WORD PTR $y1lx4+2,si	; dvy+dvx: Immed
	mov	si,cs:va		; va: Rsi
	jmp	$y1ll0
$yxneg:
	mov	cs:WORD PTR $y2lx4+2,si	; dvy+dvx: Immed
	mov	si,cs:va		; va: Rsi
	jmp	$y2ll0
$yyneg:
	or	dx,dx
	js	$yxneg1
	mov	cs:WORD PTR $y3lx4+2,si	; dvy+dvx: Immed
	mov	si,cs:va		; va: Rsi
	jmp	$y3ll0
$yxneg1:
	mov	cs:WORD PTR $y4lx4+2,si	; dvy+dvx: Immed
	mov	si,cs:va		; va: Rsi
	jmp	$y4ll0
;
; the main drawing loop (long line)
;
;
; dvx > 0  dvx+dvy > 0
;								386	486
;								===	===
$y1ll3a:
	add	si,dx			; va += dvx		  2	  1
	jc	$y1lp1			; if page crossing	  3/[7]	  1/[3]
$y1ll3:
$y1xxx:	mov	es:[si],ah		; vram = color (^=,|=)	4	2
$y1ll0:
	dec	cx			;			2	1
	js	$ylret			; if end of line	3/[7]	1/[3]
	add	bx,di			; err += dy		2	1
	jnc	$y1ll3a			; if err < dx		3/7	1/3
$y1lx4:	add	si,01234H		; va += dvx+dvy		2	1
	jnc	$y1ll3			; if NO page crossing	7/[3]	3/[1]
;								=====	=====
;								23/23	10/10
	inc	al			; ++ page
	call	setbank
	jmp	SHORT $y1ll3
$y1lp1:
	inc	al			; ++ page
	call	setbank
	jmp	SHORT $y1ll3
;
;
; dvx < 0  dvx+dvy > 0
;								386	486
;								===	===
$y2ll3a:
	add	si,dx			; va += dvx		  2	  1
	jnc	$y2lp1			; if page crossing	  3/[7]	  1/[3]
$y2ll3:
$y2xxx:	mov	es:[si],ah		; vram = color (^=,|=)	4	2
$y2ll0:
	dec	cx			;			2	1
	js	$ylret			; if end of line	3/[7]	1/[3]
	add	bx,di			; err += dy		2	1
	jnc	$y2ll3a			; if err < dx		3/7	1/3
$y2lx4:	add	si,01234H		; va += dvx+dvy		2	1
	jnc	$y2ll3			; if NO page crossing	7/[3]	3/[1]
;								=====	=====
;								23/23	10/10
	inc	al			; ++ page
	call	setbank
	jmp	SHORT $y2ll3
$y2lp1:
	dec	al			; ++ page
	call	setbank
	jmp	SHORT $y2ll3
;
;
; dvx > 0  dvx+dvy < 0
;								386	486
;								===	===
$y3ll3a:
	add	si,dx			; va += dvx		  2	  1
	jc	$y3lp1			; if page crossing	  3/[7]	  1/[3]
$y3ll3:
$y3xxx:	mov	es:[si],ah		; vram = color (^=,|=)	4	2
$y3ll0:
	dec	cx			;			2	1
	js	$ylret			; if end of line	3/[7]	1/[3]
	add	bx,di			; err += dy		2	1
	jnc	$y3ll3a			; if err < dx		3/7	1/3
$y3lx4:	add	si,01234H		; va += dvx+dvy		2	1
	jc	$y3ll3			; if NO page crossing	7/[3]	3/[1]
;								=====	=====
;								23/23	10/10
	dec	al			; ++ page
	call	setbank
	jmp	SHORT $y3ll3
$y3lp1:
	inc	al			; ++ page
	call	setbank
	jmp	SHORT $y3ll3
;
;
; dvx < 0  dvx+dvy < 0
;								386	486
;								===	===
$y4ll3a:
	add	si,dx			; va += dvx		  2	  1
	jnc	$y4lp1			; if page crossing	  3/[7]	  1/[3]
$y4ll3:
$y4xxx:	mov	es:[si],ah		; vram = color (^=,|=)	4	2
$y4ll0:
	dec	cx			;			2	1
	js	$ylret			; if end of line	3/[7]	1/[3]
	add	bx,di			; err += dy		2	1
	jnc	$y4ll3a			; if err < dx		3/7	1/3
$y4lx4:	add	si,01234H		; va += dvx+dvy		2	1
	jc	$y4ll3			; if NO page crossing	7/[3]	3/[1]
;								=====	=====
;								23/23	10/10
	dec	al			; ++ page
	call	setbank
	jmp	SHORT $y4ll3
$y4lp1:
	dec	al			; ++ page
	call	setbank
	jmp	SHORT $y4ll3

$ylret:
	jmp	GrDTx2

;							; ******************
; Short line version.					; *     NOT USED   *
;							; ******************
$xloop:
	mov	cs:WORD PTR $xlx8+2,dx	; 2  1 (old)	; dy: Immed
	mov	di,cx			; 2  1 (old)	; dx: Rdi
;					;== ==
;					; 4  2 (old)
;
	mov	dx,bx			; dvx: Rdx
	or	dx,dx			; pdvx ?
	mov	bx,07204H		; jc:add
	jns	$xlpdvx
	mov	bx,0732cH		; jnc:sub
$xlpdvx:
	mov	cs:BYTE PTR $xlx2,bh	; set jc/jnc
	mov	cs:BYTE PTR $xlx3,bl	; set add/sub
;
	mov	ax,si
	add	ax,dx
	mov	cs:WORD PTR $xlx4+2,ax	; dvx+dvy: Immed
;	or	ax,ax			; pdvy ?
	mov	bx,07304H		; jnc:add
	jns	$xlpdvy
	mov	bx,0722cH		; jc:sub
$xlpdvy:
	mov	cs:BYTE PTR $xlx5,bh	; set jnc/jc
	mov	cs:BYTE PTR $xlx6,bl	; set add/sub
;
	mov	ax,__WritePage		; ppage: Ral
;
	mov	bx,dtc
	mov	ah,bl			; color: Rah
;
	mov	bx,cx
	sar	bx,1			; err: Rbx
;
	mov	si,cs:va		; va: Rsi
	jmp	SHORT $xll0
;
; the main drawing loop (short line)
;
;				  386  486
;				  ===  ===
$xll3a:
	add	si,dx		;   2    1	; va += dvx
$xlx2:	jc	$xlp1		;   3\7  1\3	; if page crossing
$xlp1r:
$xlp2r:
$xll3:
$xxxx1:	mov	es:[si],ah	; 4    2	; vram = color (or ^= or |=)
$xll0:
	dec	cx		; 2    1
	js	$ylret		; 3\7  1\3	; if end of line
$xlx8:	add	bx,01234H	; 2    1 	; err += dy
	cmp	bx,di		; 2    1 	; err : dx
	jl	$xll3a		; 3/7  1/3	; err < dx
	sub	bx,di		; 2    1	; err -= dx
$xlx4:	add	si,01234H	; 2    1	; va += dvx+dvy
$xlx5:	jnc	$xll3		; 7\3  3\1	; if no page crossing
;				;===== =====
;				;27/25 12/11	; total
$xlp2:
$xlx6:	add	al,001H				; ++ page (or --)
	call	setbank
	jmp	SHORT $xlp2r
$xlp1:
$xlx3:	add	al,001H				; ++ page (or --)
	call	setbank
	jmp	SHORT $xlp1r

_GrDrawToA ENDP


;-----------------------------------------------------------------------------
;
; void far GrLineA (Uint xa, Uint ya, Uint xb, Uint yb, Uint color);
;
;-----------------------------------------------------------------------------

	PUBLIC	_GrLineA
_GrLineA PROC far
ltxa	EQU	WORD PTR[BP+6]
ltya	EQU	WORD PTR[BP+8]
ltxb	EQU	WORD PTR[BP+10]
ltyb	EQU	WORD PTR[BP+12]
ltc	EQU	WORD PTR[BP+14]

	push	bp
	mov	bp,sp

	mov	ax,ltxa
	mov	bx,ltya
	cmp	ax,cs:x1
	jne	GrLt1
	cmp	bx,cs:y1
	je	GrLt2
GrLt1:
	mov	cs:x1,ax
	mov	cs:y1,bx
	mov	cs:xydone,0
GrLt2:
	push	ltxb
	push	ltyb
	push	ltc
	call	_GrDrawToA

	mov	sp,bp
	pop	bp
	ret

_GrLineA ENDP


	PUBLIC @_muldiv
@_muldiv PROC  far
;
; int far _fastcall muldiv (int x, int y, int z); [ax, dx, bx]
;
;    compute (x*y)/z
;
	imul	dx
	idiv	bx
;
	ret
@_muldiv ENDP

	PUBLIC __muldiv
__muldiv PROC  far
;
; int far muldiv (int x, int y, int z);
;
;    compute (x*y)/z
;
px	EQU	ss:WORD PTR[BX+4]
py	EQU	ss:WORD PTR[BX+6]
pz	EQU	ss:WORD PTR[BX+8]
;
	mov	bx,sp
;
	mov	ax,px
	imul	py
	idiv	pz
;
	ret
__muldiv ENDP

_TEXT	ENDS
	END
